//
//  KSDynamicLinker.c
//
//  Created by Karl Stenerud on 2013-10-02.
//
//  Copyright (c) 2012 Karl Stenerud. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall remain in place
// in this source code.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "BSG_KSDynamicLinker.h"
#include "BSG_KSArchSpecific.h"
#include "BSG_KSMachHeaders.h"

#include <limits.h>
#include <mach-o/nlist.h>
#include <string.h>

uint32_t bsg_ksdlimageNamed(const char *const imageName, bool exactMatch) {
    if (imageName != NULL) {
        const uint32_t imageCount = bsg_dyld_image_count();
        
        for (uint32_t iImg = 0; iImg < imageCount; iImg++) {
            const char *name = bsg_dyld_get_image_name(iImg);
            if (name == NULL) {
                continue; // name is null if the index is out of range per dyld(3)
            } else if (exactMatch) {
                if (strcmp(name, imageName) == 0) {
                    return iImg;
                }
            } else {
                if (strstr(name, imageName) != NULL) {
                    return iImg;
                }
            }
        }
    }
    return UINT32_MAX;
}

const uint8_t *bsg_ksdlimageUUID(const char *const imageName, bool exactMatch) {
    if (imageName != NULL) {
        const uint32_t iImg = bsg_ksdlimageNamed(imageName, exactMatch);
        if (iImg != UINT32_MAX) {
            const struct mach_header *header =  bsg_dyld_get_image_header(iImg);
            if (header != NULL) {
                uintptr_t cmdPtr = bsg_ksdlfirstCmdAfterHeader(header);
                if (cmdPtr != 0) {
                    for (uint32_t iCmd = 0; iCmd < header->ncmds; iCmd++) {
                        const struct load_command *loadCmd =
                            (struct load_command *)cmdPtr;
                        if (loadCmd->cmd == LC_UUID) {
                            struct uuid_command *uuidCmd =
                                (struct uuid_command *)cmdPtr;
                            return uuidCmd->uuid;
                        }
                        cmdPtr += loadCmd->cmdsize;
                    }
                }
            }
        }
    }
    return NULL;
}

uintptr_t bsg_ksdlfirstCmdAfterHeader(const struct mach_header *const header) {
    if (header == NULL) {
      return 0;
    }
    switch (header->magic) {
    case MH_MAGIC:
    case MH_CIGAM:
        return (uintptr_t)(header + 1);
    case MH_MAGIC_64:
    case MH_CIGAM_64:
        return (uintptr_t)(((struct mach_header_64 *)header) + 1);
    default:
        // Header is corrupt
        return 0;
    }
}

uint32_t bsg_ksdlimageIndexContainingAddress(const uintptr_t address) {
    const uint32_t imageCount = bsg_dyld_image_count();
    const struct mach_header *header = 0;

    for (uint32_t iImg = 0; iImg < imageCount; iImg++) {
        header = bsg_dyld_get_image_header(iImg);
        if (header != NULL) {
            // Look for a segment command with this address within its range.
            uintptr_t addressWSlide =
                address - bsg_dyld_get_image_vmaddr_slide(iImg);
            uintptr_t cmdPtr = bsg_ksdlfirstCmdAfterHeader(header);
            if (cmdPtr == 0) {
                continue;
            }
            for (uint32_t iCmd = 0; iCmd < header->ncmds; iCmd++) {
                const struct load_command *loadCmd =
                    (struct load_command *)cmdPtr;
                if (loadCmd->cmd == LC_SEGMENT) {
                    const struct segment_command *segCmd =
                        (struct segment_command *)cmdPtr;
                    if (addressWSlide >= segCmd->vmaddr &&
                        addressWSlide < segCmd->vmaddr + segCmd->vmsize) {
                        return iImg;
                    }
                } else if (loadCmd->cmd == LC_SEGMENT_64) {
                    const struct segment_command_64 *segCmd =
                        (struct segment_command_64 *)cmdPtr;
                    if (addressWSlide >= segCmd->vmaddr &&
                        addressWSlide < segCmd->vmaddr + segCmd->vmsize) {
                        return iImg;
                    }
                }
                cmdPtr += loadCmd->cmdsize;
            }
        }
    }
    return UINT_MAX;
}

uintptr_t bsg_ksdlsegmentBaseOfImageIndex(const uint32_t idx) {
    const struct mach_header *header =  bsg_dyld_get_image_header(idx);
    if (header == NULL) {
        return 0;
    }

    // Look for a segment command and return the file image address.
    uintptr_t cmdPtr = bsg_ksdlfirstCmdAfterHeader(header);
    if (cmdPtr == 0) {
        return 0;
    }
    for (uint32_t i = 0; i < header->ncmds; i++) {
        const struct load_command *loadCmd = (struct load_command *)cmdPtr;
        if (loadCmd->cmd == LC_SEGMENT) {
            const struct segment_command *segmentCmd =
                (struct segment_command *)cmdPtr;
            if (strcmp(segmentCmd->segname, SEG_LINKEDIT) == 0) {
                return segmentCmd->vmaddr - segmentCmd->fileoff;
            }
        } else if (loadCmd->cmd == LC_SEGMENT_64) {
            const struct segment_command_64 *segmentCmd =
                (struct segment_command_64 *)cmdPtr;
            if (strcmp(segmentCmd->segname, SEG_LINKEDIT) == 0) {
                return (uintptr_t)(segmentCmd->vmaddr - segmentCmd->fileoff);
            }
        }
        cmdPtr += loadCmd->cmdsize;
    }

    return 0;
}

bool bsg_ksdldladdr(const uintptr_t address, Dl_info *const info) {
    info->dli_fname = NULL;
    info->dli_fbase = NULL;
    info->dli_sname = NULL;
    info->dli_saddr = NULL;

    const uint32_t idx = bsg_ksdlimageIndexContainingAddress(address);
    if (idx == UINT_MAX) {
        return false;
    }
    const struct mach_header *header = bsg_dyld_get_image_header(idx);
    if (header == NULL) {
        return false;
    }
    const uintptr_t imageVMAddrSlide =
        bsg_dyld_get_image_vmaddr_slide(idx);
    const uintptr_t addressWithSlide = address - imageVMAddrSlide;
    const uintptr_t segmentBase =
        bsg_ksdlsegmentBaseOfImageIndex(idx) + imageVMAddrSlide;
    if (segmentBase == 0) {
        return false;
    }

    info->dli_fname = bsg_dyld_get_image_name(idx);
    info->dli_fbase = (void *)header;

    // Find symbol tables and get whichever symbol is closest to the address.
    const BSG_STRUCT_NLIST *bestMatch = NULL;
    uintptr_t bestDistance = ULONG_MAX;
    uintptr_t cmdPtr = bsg_ksdlfirstCmdAfterHeader(header);
    if (cmdPtr == 0) {
        return false;
    }
    for (uint32_t iCmd = 0; iCmd < header->ncmds; iCmd++) {
        const struct load_command *loadCmd = (struct load_command *)cmdPtr;
        if (loadCmd->cmd == LC_SYMTAB) {
            const struct symtab_command *symtabCmd =
                (struct symtab_command *)cmdPtr;
            const BSG_STRUCT_NLIST *symbolTable =
                (BSG_STRUCT_NLIST *)(segmentBase + symtabCmd->symoff);
            const uintptr_t stringTable = segmentBase + symtabCmd->stroff;

            for (uint32_t iSym = 0; iSym < symtabCmd->nsyms; iSym++) {
                // If n_value is 0, the symbol refers to an external object.
                if (symbolTable[iSym].n_value != 0) {
                    uintptr_t symbolBase = symbolTable[iSym].n_value;
                    uintptr_t currentDistance = addressWithSlide - symbolBase;
                    if ((addressWithSlide >= symbolBase) &&
                        (currentDistance <= bestDistance)) {
                        bestMatch = symbolTable + iSym;
                        bestDistance = currentDistance;
                    }
                }
            }
            if (bestMatch != NULL) {
                info->dli_saddr =
                    (void *)(bestMatch->n_value + imageVMAddrSlide);
                info->dli_sname = (char *)((intptr_t)stringTable +
                                           (intptr_t)bestMatch->n_un.n_strx);
                if (*info->dli_sname == '_') {
                    info->dli_sname++;
                }
                // This happens if all symbols have been stripped.
                if (info->dli_saddr == info->dli_fbase &&
                    bestMatch->n_type == 3) {
                    info->dli_sname = NULL;
                }
                break;
            }
        }
        cmdPtr += loadCmd->cmdsize;
    }

    return true;
}

const void *bsg_ksdlgetSymbolAddrInImage(uint32_t imageIdx,
                                         const char *symbolName) {
    const struct mach_header *header = bsg_dyld_get_image_header(imageIdx);
    if (header == NULL) {
        return NULL;
    }
    const uintptr_t imageVMAddrSlide =
        bsg_dyld_get_image_vmaddr_slide(imageIdx);
    const uintptr_t segmentBase =
        bsg_ksdlsegmentBaseOfImageIndex(imageIdx) + imageVMAddrSlide;
    if (segmentBase == 0) {
        return NULL;
    }
    uintptr_t cmdPtr = bsg_ksdlfirstCmdAfterHeader(header);
    if (cmdPtr == 0) {
        return NULL;
    }
    for (uint32_t iCmd = 0; iCmd < header->ncmds; iCmd++) {
        const struct load_command *loadCmd = (struct load_command *)cmdPtr;
        if (loadCmd->cmd == LC_SYMTAB) {
            const struct symtab_command *symtabCmd =
                (struct symtab_command *)cmdPtr;
            const BSG_STRUCT_NLIST *symbolTable =
                (BSG_STRUCT_NLIST *)(segmentBase + symtabCmd->symoff);
            const uintptr_t stringTable = segmentBase + symtabCmd->stroff;

            for (uint32_t iSym = 0; iSym < symtabCmd->nsyms; iSym++) {
                // If n_value is 0, the symbol refers to an external object.
                if (symbolTable[iSym].n_value != 0) {
                    const char *sname =
                        (char *)((intptr_t)stringTable +
                                 (intptr_t)symbolTable[iSym].n_un.n_strx);
                    if (*sname == '_') {
                        sname++;
                    }
                    if (strcmp(sname, symbolName) == 0) {
                        return (void *)(symbolTable[iSym].n_value +
                                        imageVMAddrSlide);
                    }
                }
            }
        }
        cmdPtr += loadCmd->cmdsize;
    }
    return NULL;
}

const void *bsg_ksdlgetSymbolAddrInAnyImage(const char *symbolName) {
    const uint32_t imageCount = bsg_dyld_image_count();

    for (uint32_t iImg = 0; iImg < imageCount; iImg++) {
        const void *symbolAddr = bsg_ksdlgetSymbolAddrInImage(iImg, symbolName);
        if (symbolAddr != NULL) {
            return symbolAddr;
        }
    }
    return NULL;
}
